#!/bin/bash
# Control script for polyphase FFT+GBM+Recovery, Florian Steinbach
# Version 150806_1112

# # INFO # # 
# Order of processes performed:
# 1. FFT (elle2fft PPC fft2elle)
# 2. Import FFT data to ellefile
# 3. Calculate DDs from Misorientation overwriting the FFT ones (if switched on)
# 4. Elle process with several topology checks (FS_topocheck)
# 5. Recovery (if switched on) 
# 6. GBM (if switched on) 
# 7. FS_flynn2unode_attribute for VISCOSITY
# 8. Create plotlayer deleting all VISC=2 (bubble) unodes and merging relevant bubble flynns  (if switched on) 

DEBUG=
#DEBUG="echo"
rm *~

#############################################################################
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#                                   INPUT                                   #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#############################################################################

INROOT=fraction00             # Input file, do not type file extension
OUTPATH="results/"          # output folder, folder "stepXXX" will be stored in here 
OUTROOT=fraction00rate10_newgrid            # Output fileroot, extension "_stepXXX.elle" will be added
TOTALSTEPS=75                # Number of steps to perform
STARTSTEP=0                 # Steps already performed in starting elle file

DIMENSIONS=256              # Number of unodes in x- and y-direction
AIRPHASE=2                  # What is the phase ID of air?! --> Only helpful for user infos in gbm
EXCLUDEDPHASE=0             # ID of phase being excluded from DD calculation in fft2elle (no exclusion: set to 0)
REPOSITION=0                # Set to 1 if reposition is needed, to 0 if not
DD_MISORI=-1                 # Set to anything else than -1 if DDs shall be calculated from misorientations
                            # Attention: The Number will indicate the order of neighbours used for KAM

LOGSCREEN=0                 # Leave at zero, only change to 1 if you wish to switch off randomisation 
GBM_STEPS=20                 # number of gbm steps (0=do not perform GBM)
REC_STEPS=20                 # number of recovery steps (0=do not perform recovery)
    ROT_MOB=500             # Give a value for rotation mobility (1/Pa s m²)
    HAGB=15                 # High angle grain boundary angle for recovery
SAVEPLOTLAYERS=1            # Set to 1 if you wish to save a plotlayer file without bubble unodes etc. otherwise set to 0
SCALEBOX4PLOT=0.8          # In pure shear models maybe useful to plot unodes without black stripes: 
                            # Type scaling exponent (<1 to scale down) if you want to use this or 
                            # type 0 if you do not want to use this
SAVEGRIDDATA=$DIMENSIONS     # Include data for passive marker grids? Type 0 if not, type the dimension if yes

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#                       USER INFO AND PREPARATIONS                          #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

echo ~~~~~~~~~~~~~~~~ STARTING ~~~~~~~~~~~~~~~~

# zip all input files to save them:
$DEBUG zip -r input_files.zip .

$DEBUG cp $INROOT.elle tmp.elle

# If the directory $OUTPATH does not exist: create it:
if [ ! -d "$OUTPATH" ]; then
    $DEBUG mkdir $OUTPATH    
fi

# If the directory $OUTPATH/plotlayers does not exist: create it if user wishes to save plotlayers
if [ $SAVEPLOTLAYERS -ne 0 ];
then
    PLOTLAYER_PATH=$OUTPATH"/plotlayers/"    
    if [ ! -d "$PLOTLAYER_PATH" ]; then
        $DEBUG mkdir $PLOTLAYER_PATH
    fi
fi


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#                         LOOP THROUGH STEPS                                #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

for (( i=1; i<=$TOTALSTEPS; i++ )) 
do	
	step=$(printf "%03d" $(($i+$STARTSTEP)) )
	echo ~~~~~~~~~~~~~~ STEP $(($i+$STARTSTEP))" of "$(($TOTALSTEPS+$STARTSTEP))  ~~~~~~~~~~~~~~~

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	#                         ELLE2FFT PROCESS                                  #
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

    printf "\n~~~~~~~~~~~~~~ ELLE2FFT ~~~~~~~~~~~~~~\n\n"
	$DEBUG FS_elle2fft -i tmp.elle -u $DIMENSIONS -n
	$DEBUG mv elle2fft.elle tmp.elle 

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	#           FFT ITSELF AND FFT2ELLE (Florian's "FS_fft2elle")                #
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
        
    printf "\n~~~~~~~~~~~~~~~~ FFT ~~~~~~~~~~~~~~~~\n\n"
	$DEBUG FFT_vs$DIMENSIONS
    printf "\n~~~~~~~~~~~~~~ FFT2ELLE ~~~~~~~~~~~~~~\n\n"
	$DEBUG FS_fft2elle -i tmp.elle -u 1 $EXCLUDEDPHASE $SAVEGRIDDATA -n
	$DEBUG mv fft2elle.elle tmp.elle

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	#                    REPOSITION (only for simple shear)                     #
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    if [ $REPOSITION -eq 1 ];
    then
        printf "\n~~~~~~~~~~~~~ REPOSITION ~~~~~~~~~~~~~\n\n"
        $DEBUG reposition -i tmp.elle -n
        $DEBUG mv repos.elle tmp.elle    
    fi

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	#                   IMPORT MORE DATA FROM tex.out FILE                      #
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    
    printf "\n~~~~~~~~~~~ IMPORT FFT DATA ~~~~~~~~~~~\n\n"
	$DEBUG importFFTdata_florian -i tmp.elle -u 4 5 6 7 12 -s 1 -f 1 -n
	$DEBUG mv fft_out.elle tmp.elle

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	#         CALCULATE DISLOATION DENSITIES FROM MISORIENTATION                #
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    if [ $DD_MISORI -ne -1 ];
    then
        printf "\n~~~~~~~~~~~~~~ MISORI2DD ~~~~~~~~~~~~~~\n\n"     
        $DEBUG cp tmp.elle before_misori2dd.elle
	    $DEBUG misori2dd -i tmp.elle -u $DD_MISORI -n
	    $DEBUG mv misori2dd001.elle tmp.elle
    fi

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	#           Topology checks (also performed after each GBM step later       #
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    printf "\n~~~~~~~~~~~ TOPOLOGY CHECKS ~~~~~~~~~~~\n\n"
	$DEBUG FS_topocheck -i tmp.elle -n
	$DEBUG mv FS_topocheck.elle tmp.elle

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	#                            RECOVERY                                       #
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    if [ $REC_STEPS -ne 0 ];
    then
        printf "\n~~~~~~~~~~~~~~~ RECOVERY ~~~~~~~~~~~~~~~\n\n" 
        $DEBUG cp tmp.elle before_recovery.elle	    
        $DEBUG FS_recovery -i tmp.elle -u $HAGB $AIRPHASE $LOGSCREEN $ROT_MOB -s $REC_STEPS -f $REC_STEPS -n
        recstep3digit=$(printf "%03d" $REC_STEPS )	
        if [ -f recovery$recstep3digit.elle ];
        then
            $DEBUG mv recovery$recstep3digit.elle tmp.elle
        else
            echo "CONTROL SCRIPT ERROR: File recovery$recstep3digit.elle is not existing, stopping the simulation at step $step"
            $DEBUG exit # not using break any more to definitely stop the script at this point
        fi
    fi

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	#                                  GBM                                      #
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    if [ $GBM_STEPS -ne 0 ];
    then
        printf "\n~~~~~~~~~~~~~~~~~ GBM ~~~~~~~~~~~~~~~~~\n\n"
        $DEBUG cp tmp.elle before_gbm.elle    
        $DEBUG FS_gbm_pp_fft -i tmp.elle -u 0 0 $LOGSCREEN 0 $AIRPHASE -s $GBM_STEPS -f $GBM_STEPS -n
        gbmstep3digit=$(printf "%03d" $GBM_STEPS )
        if [ -f gbm_pp_fft$gbmstep3digit.elle ];
        then
            $DEBUG mv gbm_pp_fft$gbmstep3digit.elle tmp.elle
        else
            echo "CONTROL SCRIPT ERROR: File gbm_pp_fft$gbmstep3digit.elle is not existing, stopping the simulation at step $step"
            $DEBUG exit # not using break any more to definitely stop the script at this point
        fi
    fi

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	#      SPECIALITY OF POLYPHASE: UPDATE PHASE ID (U_VISCOSITY) IN UNODES     #
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    $DEBUG FS_flynn2unode_attribute -i tmp.elle -u 1 -n
    $DEBUG mv flynn2unode_attribute.elle tmp.elle	
    
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	#   SPECIALITY OF POLYPHASE: CREATE A PLOTLAYER FILE WITHOUT BUBBLE UNODES  #
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    
    if [ $SAVEPLOTLAYERS -ne 0 ];
    then
        printf "\n~~~~~~~~~ CREATE PLOTTING FILE ~~~~~~~~~\n\n"
        $DEBUG FS_create_plotlayer -i tmp.elle -u $AIRPHASE $AIRPHASE $SCALEBOX4PLOT 0 0 -n
        $DEBUG mv with_plotlayer.elle tmp_plotlayer.elle
    fi
        
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	#                   FINALLY RENAME AND SAVE THE FILES                       #
	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    
    # Save final ellefile
    printf "\n~~~~~~~~~~~ STORE ALL FILES ~~~~~~~~~~~\n\n"
    $DEBUG cp tmp.elle $OUTPATH$OUTROOT"_step"$step".elle"
    stepdirectory=$OUTPATH"step"$step"/"

    # Save plotlayer file without bubble unodes if user switched that option on:
    if [ $SAVEPLOTLAYERS -ne 0 ];
    then
        $DEBUG mv tmp_plotlayer.elle $PLOTLAYER_PATH$OUTROOT"_plotlayer_step"$step".elle"
    fi

    # Save other files
    $DEBUG mkdir $stepdirectory
	$DEBUG cp g0.out $stepdirectory"g0_"$step.out 
	$DEBUG cp tex.out $stepdirectory"tex"$step.out 
	$DEBUG cp all.out $stepdirectory"all"$step.out
	$DEBUG cp conv.out $stepdirectory"conv"$step.out
	$DEBUG cp err.out $stepdirectory"err"$step.out
	$DEBUG cp temp.out $stepdirectory"temp"$step.out
	$DEBUG cp temp-FFT.out $stepdirectory"temp-FFT"$step.out
	$DEBUG cp unodexyz.out $stepdirectory"unodexyz"$step.out
	$DEBUG cp unodeang.out $stepdirectory"unodeang"$step.out
	#$DEBUG cp eunodes.data $stepdirectory"eunodes"$step.data
	$DEBUG mv before_gbm.elle $stepdirectory"before_gbm"$step.elle
	$DEBUG mv before_recovery.elle $stepdirectory"before_recovery"$step.elle
	$DEBUG mv before_misori2dd.elle $stepdirectory"before_misori2dd"$step.elle
    $DEBUG mv recovery.stats $stepdirectory"recovery"$step.stats
    $DEBUG mv Logfile* $stepdirectory

    # uncomment the following five lines to zip the results additionally:
    #$DEBUG cd $OUTPATH
    #$DEBUG zip step$step.zip $OUTROOT"_step"$step".elle" step$step/*
    #$DEBUG rm -rf step$step
    #$DEBUG rm $OUTROOT"_step"$step".elle"     
    #$DEBUG cd ..
    
done

if [ $GBM_STEPS -ne 0 ];
then
    $DEBUG mkdir $OUTPATH"files_gbm"/
    $DEBUG mv initial_stuff.txt $OUTPATH"files_gbm"/"initial_stuff.txt"
    $DEBUG mv PhaseAreaHistory-txt $OUTPATH"files_gbm"/
fi
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#                               ZIP FILES                                   #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
cd $OUTPATH
    $DEBUG zip -r "ALL_RESULTS.zip" .
cd ..

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#             DELETE FILES, THEY HAVE BEEN STORED ELSEWHERE                 #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

$DEBUG rm *.unodes *.flynn *.stats *.out tmp.elle *.data
echo " "
echo " "
echo ~~~~~~~~~~~~~~~~ FINISHED ~~~~~~~~~~~~~~~~ 
